use bus::Bus;
use crate::{
    app::{
        App,
        authentication::Authentication,
        user::User,
        connection::Connection,
    },
    protocol,
};

pub async fn thread_system(mut app:App, bus:Bus<protocol::QRPacket>)
{
    use protocol::*;
    use protocol::QRPacketData::*;
    use ring::rand::{SecureRandom, SystemRandom};

    let rng = SystemRandom::new();
    let argon_config = argon2::Config::default();

    while match bus.receive_wait() {
        Some(packet) => {
            let qr = &packet.data;
            match &qr.data {
                QConn(request) => {
                    let id = app.connections.add(Connection {
                        bus: request.bus_id,
                        auth: None,
                    });

                    println!("Connect: {}", id);

                    bus.send(packet.from, QRPacket::new(id as u32, RConn)).is_ok()
                }

                QDisconn => {
                    app.connections.remove(qr.id as usize).ok();

                    println!("Disconnect: {}", qr.id);
                    true
                }

                QRegister(request) => {
                    let mut response = PacketRegisterResponse::new();
                    response.status = STATUS_ERROR;

                    println!("Request: Register");

                    let mut is_valid = true;
                    if request.code != "abc".as_bytes() { response.status = STATUS_BAD_CODE; is_valid = false; }
                    if is_valid && request.handle.len() == 0 { response.status = STATUS_BAD_HANDLE; is_valid = false; }
                    if is_valid && request.secret.len() == 0 { response.status = STATUS_BAD_SECRET; is_valid = false; }

                    if is_valid {
                        match app.user_handle.get(request.handle.as_bytes()) {
                            None => {
                                let mut salt = [0u8; 16];
                                match rng.fill(&mut salt) {
                                    Ok(_) => {
                                        let salt_id = app.salts.len();
                                        app.salts.push(salt);

                                        if let Ok(hash) = argon2::hash_raw(&request.secret, &salt, &argon_config) {
                                            let user_id = app.user_next;
                                            app.user_next += 1;

                                            // Create user entry
                                            let user_pos = app.users.add(User {
                                                id:user_id,
                                                handle:request.handle.clone(),
                                                secret:hash,
                                                na_key:salt_id,
                                            });

                                            // Register user pool id and handle
                                            app.user_id.set(user_id as isize, user_pos);
                                            app.user_handle.set(request.handle.as_bytes(), user_id);

                                            println!("Registered user '{}' @ {} with id {}", request.handle, user_pos, user_id);

                                            // Generate authentication token and secret
                                            response.status = STATUS_OK;
                                            rng.fill(&mut response.secret).ok();
                                            loop {
                                                rng.fill(&mut response.token).ok();

                                                if app.auths.get(&response.token).is_none() {
                                                    app.auths.set(&response.token, Authentication {
                                                        key:response.token,
                                                        secret:response.secret,
                                                        user:user_id,
                                                    });
                                                    break;
                                                }
                                            }
                                            
                                            // Attach authentication to connection
                                            if let Some(conn) = app.connections.get_mut(qr.id as usize) {
                                                conn.auth = Some(response.token);
                                            }
                                        }
                                    }
                                    Err(_) => { println!("error: failed to generate salt.") }
                                }
                            }
                            Some(_) => {
                                response.status = STATUS_BAD_HANDLE;
                                println!("notice: attempt to register existing handle: '{}'", request.handle);
                            }
                        }
                    }

                    bus.send(packet.from, QRPacket::new(qr.id, RRegister(response))).is_ok()
                }

                QAuth(request) => {
                    let mut response = PacketAuthResponse::new();
                    response.status = STATUS_ERROR;

                    println!("Request: Auth");

                    let mut is_valid = true;
                    if is_valid && request.handle.len() == 0 { response.status = STATUS_BAD_HANDLE; is_valid = false; }
                    if is_valid && request.secret.len() == 0 { response.status = STATUS_BAD_SECRET; is_valid = false; }

                    if is_valid {

                        // Get user data from handle
                        match app.user_handle.get(request.handle.as_bytes()) {
                            Some(uid) => {
                                if let Some(tuid) = app.user_id.get(*uid as isize) {
                                    if let Some(user) = app.users.get(*tuid) {
                                        // Get user salt
                                        if let Some(salt) = app.salts.get(user.na_key) {

                                            // Verify salted secret against user data
                                            if argon2::verify_raw(&request.secret.as_bytes(), salt, &user.secret, &argon_config).unwrap_or(false) {
                                                println!("Authenticated user '{}' id {}", request.handle, *uid);

                                                // Generate authentication token and secret
                                                response.status = STATUS_OK;
                                                rng.fill(&mut response.secret).ok();
                                                loop {
                                                    rng.fill(&mut response.token).ok();

                                                    if app.auths.get(&response.token).is_none() {
                                                        app.auths.set(&response.token, Authentication {
                                                            key:response.token,
                                                            secret:response.secret,
                                                            user:*uid,
                                                        });
                                                        break;
                                                    }
                                                }
                                                
                                                // Attach authentication to connection
                                                if let Some(conn) = app.connections.get_mut(qr.id as usize) {
                                                    conn.auth = Some(response.token);
                                                }
                                            } else {
                                                println!("notice: password verification failed.");
                                            }
                                        } else {
                                            println!("error: user salt id '{}' not found.", user.na_key);
                                        }
                                    } else {
                                        println!("error: user with id '{}' not found.", uid);
                                    }
                                } else {
                                    println!("error: user with id '{}' not found.", uid);
                                }
                            }
                            None => { }
                        }
                    }

                    bus.send(packet.from, QRPacket::new(qr.id, RAuth(response))).is_ok()
                }

                QDeauth => {
                    if let Some(conn) = app.connections.get_mut(qr.id as usize) {
                        match conn.auth {
                            Some(auth) => {
                                println!("Deauthenticated connection: {}", qr.id);
                                app.auths.unset(&auth);
                            }
                            None => { }
                        }
                        conn.auth = None;
                    }
                    true
                }

                _ => { true }
            }
        }
        None => false,
    } { }
}
